unit ExcelPvd;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
  Windows, SysUtils, db,  Variants, Classes,ArtRepProvider;

                         
type TARExcelProvider = class(TArtRepProvider)
 private
   EA : variant;
   WBk: variant;
   WS : variant;
   flocal : boolean;
   function ProgLoaded:boolean;
 public
    procedure Connect; override;
    procedure SetVisible(visible:boolean); override;
    procedure OpenDoc(name:string; page:integer); override;
    procedure CreateDoc(name:string); override;
    procedure SetCalculation(auto:boolean); override;
    function GetCellText(row,col:integer) : string; override;
    procedure SetCellText(row,col:integer; txt:string); override;
    function GetCellFormula(row,col:integer) : string; override;
    procedure SetCellFormula(row,col:integer; Formula:Variant); override;
    procedure DeleteRow(row : integer); override;
    procedure DeleteRows(row, count : integer); override;
    procedure DeleteColumn(col : integer); override;
    procedure InsertColumns(col, count:integer); override;
    procedure CopyColumns(col, tocol, count:integer); override;
    procedure InsertRows(row, count:integer); override;
    procedure CopyRows(row, torow, count:integer); override;
    procedure SetRange(R1,C1,R2,C2:integer; Data:Variant); override;
    procedure AddNameRange(R1,C1,R2,C2:integer; name:string); override;
    procedure AddName(name:string;val:integer); override;
    procedure SetSheetName(name : string); override;
    function LastUsedRow : integer; override;
    function DocLoaded : boolean; override;
    function LastUsedColumn : integer; override;
    procedure Group(row, count : integer); override;
    procedure CleanUp; override;
    procedure AutoFit(col : integer); override;
    procedure ActivateCell(row,col : integer); override;
    procedure ShowLevels(level : integer); override;
    procedure Run(macro : string); override;
    procedure Print; override;
    procedure SaveDocAs(s:string); override;
    procedure SaveDoc; override;
    procedure BeforeQuit; override;
    procedure CloseDoc; override;
    procedure MarkSaved; override;
    procedure Quit; override;
    function Installed : boolean; override;
    function ProviderName : string; override;
    function GetMyVariant(Field: TField;str:boolean=true): OLEVariant; override;
    function GetDate(val : TDateTime): OLEVariant; override;
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
  published
    property FormulaLocal : boolean read flocal write flocal default false;   
end;

implementation

uses oleserver, activex, comobj;

const

 {$IFDEF FPC}
 xlCalculationAutomatic = -4105;
 xlCalculationManual = -4135;
 xlCalculationSemiautomatic = 2;
 xlShiftDown = -4121;
 xlShiftToRight = -4161;
 xlMaximized = -4137;
 xlMinimized = -4140;
 xlNormal = -4143;
 xlExclusive = 3;
 xlNoChange = 1;
 xlShared = 2;
 xlLocalSessionChanges = 2;
 xlOtherSessionChanges = 3;
 xlUserResolution = 1;
 {$ELSE}
  xlCalculationAutomatic = $FFFFEFF7;
  xlCalculationManual = $FFFFEFD9;
  xlCalculationSemiautomatic = $00000002;
  xlShiftDown = $FFFFEFE7;
  xlShiftToRight = $FFFFEFBF;
  xlMaximized = $FFFFEFD7;
  xlMinimized = $FFFFEFD4;
  xlNormal = $FFFFEFD1;
  xlExclusive = $00000003;
  xlNoChange = $00000001;
  xlShared = $00000002;
  xlLocalSessionChanges = $00000002;
  xlOtherSessionChanges = $00000003;
  xlUserResolution = $00000001;
 {$ENDIF}

constructor TARExcelProvider.Create(Owner: TComponent);
begin
  inherited Create(Owner);
//  lcid := LOCALE_USER_DEFAULT;//GetUserDefaultLCID;
end;

destructor TARExcelProvider.Destroy;
begin
  inherited Destroy;
end;

function TARExcelProvider.ProgLoaded : boolean;
begin
 result := not (VarIsEmpty(EA) or VarIsNull(EA));
end;


procedure TARExcelProvider.Connect;
begin
 if not ProgLoaded then
   EA := CreateOleObject('Excel.Application');
end;

procedure TARExcelProvider.SetSheetName(name : string);
begin
 ws.Name := {$IFDEF FPC}widestring({$ENDIF}name{$IFDEF FPC}){$ENDIF};
end;

function TARExcelProvider.Installed : boolean;
var ClassID: TCLSID;
begin
  result := CLSIDFromProgID(PWideChar(WideString('Excel.Application')), ClassID) = S_OK;
end;

function TARExcelProvider.ProviderName : string;
begin
 result := 'MS Excel';
end;

procedure TARExcelProvider.SetVisible(visible:boolean);
begin
   ea.ScreenUpdating := visible;
   ea.EnableEvents := visible;
   ea.Visible := visible;
end;

procedure TARExcelProvider.OpenDoc(name:string; page:integer);
begin

   wbk := ea.Workbooks.open( {$IFDEF FPC}widestring(name){$ELSE}name{$ENDIF},
                emptyparam,emptyparam,emptyparam,
                emptyparam,emptyparam,emptyparam,emptyparam,
                emptyparam,emptyparam,emptyparam,emptyparam,
                emptyparam);
   if page > 0 then WS := Wbk.Sheets.item[page]
   else WS := WBk.ActiveSheet;
   WS.Activate;
end;

procedure TARExcelProvider.CreateDoc(name:string);
begin
 {$IFDEF FPC}
 wbk := ea.Workbooks.Add(widestring(name));
 {$ELSE}
 wbk := ea.Workbooks.Add(name);
 {$ENDIF}
   WS := WBk.ActiveSheet;
end;

procedure TARExcelProvider.SetCalculation(auto:boolean);
begin
  if not auto then
   ea.Calculation := xlCalculationManual
  else
    ea.Calculation := xlCalculationAutomatic;

end;

function TARExcelProvider.GetCellText(row,col:integer) : string;
begin
 result := ws.Cells.item[row,col].text;
end;

Procedure TARExcelProvider.SetCellText(row,col:integer; txt:string);
begin
 ws.Cells.Item[row,col].value := {$IFDEF FPC}widestring({$ENDIF}txt{$IFDEF FPC}){$ENDIF};
end;

function TARExcelProvider.GetCellFormula(row,col:integer) : string;
begin
 if flocal then
  result := ws.Cells.Item[row,col].FormulaR1C1Local
 else
  result := ws.Cells.Item[row,col].FormulaR1C1;
end;

Procedure TARExcelProvider.SetCellFormula(row,col:integer; Formula:Variant);
begin
 if flocal then
  ws.Cells.Item[row,col].FormulaR1C1Local := Formula
 else
  ws.Cells.Item[row,col].FormulaR1C1 := Formula
end;


procedure TARExcelProvider.DeleteRow(row : integer);
begin
 ws.Cells.item[row,1].EntireRow.delete;
end;

procedure TARExcelProvider.DeleteColumn(col : integer);
begin
  ws.Cells.Item[1,col].EntireColumn.Delete;
end;

procedure TARExcelProvider.InsertColumns(col, count:integer);
begin
 {$IFDEF FPC}
// ws.range[widestring(colname(col)+'1'),widestring(colname(col+count-1)+'1')].Entirecolumn.Insert(xlshifttoright);
 ws.range[ws.columns.item[col],ws.columns.item[col+count-1]].Entirecolumn.Insert(xlshifttoright);
 {$ELSE}
 ws.range[colname(col)+'1',colname(col+count-1)+'1'].Entirecolumn.Insert(xlshifttoright);
 {$ENDIF}
end;

procedure TARExcelProvider.CopyColumns(col, tocol, count:integer);
begin
 {$IFDEF FPC}
// ws.range[widestring(colname(col)+'1'),widestring(colname(col+count-1)+'1')].
//     EntireColumn.copy(ws.range[widestring(colname(tocol)+'1'),widestring(colname(tocol+count-1)+'1')].Entirecolumn);
 ws.range[ws.columns.item[col],ws.columns.item[col+count-1]].
  EntireColumn.copy(ws.range[ws.columns.item[tocol],ws.columns.item[tocol+count-1]].Entirecolumn);
 {$ELSE}
 ws.range[colname(col)+'1',colname(col+count-1)+'1'].EntireColumn.copy(ws.range[colname(tocol)+'1',colname(tocol+count-1)+'1'].Entirecolumn);
 {$ENDIF}
end;


procedure TARExcelProvider.InsertRows(row, count:integer);
begin
 {$IFDEF FPC}
// ws.range[widestring('A'+inttostr(row)),widestring('A'+inttostr(row+count-1))].EntireRow.Insert(xlshiftDown);
  ws.range[ws.rows.item[row],ws.rows.item[row+count-1]].EntireRow.Insert(xlshiftDown);
 {$ELSE}
 ws.range['A'+inttostr(row),'A'+inttostr(row+count-1)].EntireRow.Insert(xlshiftDown);
 {$ENDIF}
end;

procedure TARExcelProvider.DeleteRows(row, count:integer);
begin
 {$IFDEF FPC}
// ws.range[widestring('A'+inttostr(row)),widestring('A'+inttostr(row+count-1))].EntireRow.delete(emptyparam);
 ws.range[ws.rows.item[row],ws.rows.item[row+count-1]].EntireRow.delete(emptyparam);
 {$ELSE}
 ws.range['A'+inttostr(row),'A'+inttostr(row+count-1)].EntireRow.delete(emptyparam);
 {$ENDIF}
end;


procedure TARExcelProvider.CopyRows(row, torow, count:integer);
begin
 {$IFDEF FPC}
// ws.range[widestring('A'+inttostr(row)),widestring('A'+inttostr(row+count-1))].
//  EntireRow.copy(ws.range[widestring('A'+inttostr(torow)),widestring('A'+inttostr(torow))].Entirerow);

 ws.range[ws.rows.item[row],ws.rows.item[row+count-1]].
  EntireRow.copy(ws.range[ws.rows.item[torow],ws.rows.item[torow+count-1]].Entirerow);

 {$ELSE}
 ws.range['A'+inttostr(row),'A'+inttostr(row+count-1)].EntireRow.copy(ws.range['A'+inttostr(torow),'A'+inttostr(torow)].Entirerow);
 {$ENDIF}
end;

procedure TARExcelProvider.SetRange(R1,C1,R2,C2:integer; Data:Variant);
begin
 if flocal then
  {$IFDEF FPC}
//  ws.Range[widestring(colname(C1)+inttostr(R1)),widestring(colname(C2)+inttostr(R2))].FormulaR1C1Local := Data
  ws.Range[ws.cells.item[R1,C1],ws.cells.item[R2,C2]].FormulaR1C1Local := Data
 else
//  ws.Range[widestring(colname(C1)+inttostr(R1)),widestring(colname(C2)+inttostr(R2))].FormulaR1C1 := Data;
  ws.Range[ws.cells.item[R1,C1],ws.cells.item[R2,C2]].FormulaR1C1 := Data;
  {$ELSE}
  ws.Range[colname(C1)+inttostr(R1),colname(C2)+inttostr(R2)].FormulaR1C1Local := Data
 else
  ws.Range[colname(C1)+inttostr(R1),colname(C2)+inttostr(R2)].FormulaR1C1 := Data;
  {$ENDIF}
end;

procedure TARExcelProvider.AddNameRange(R1,C1,R2,C2:integer; name:string); 
var n,v : widestring;
begin
 n := ''''+ws.Name+'''!'+name;
 v := '='''+ws.Name+'''!R'+inttostr(R1)+'C'+inttostr(C1)+':R'+inttostr(R2)+'C'+inttostr(C2);
 wbk.Names.Add(n,EmptyParam,EmptyParam,EmptyParam,EmptyParam,
         EmptyParam,EmptyParam,EmptyParam,EmptyParam,
         v,EmptyParam);
end;


procedure TARExcelProvider.AddName(name:string;val:integer);
var n,v : widestring;
begin
 n := ''''+ws.Name+'''!'+name;
 v := '='+inttostr(val);
 wbk.Names.Add(n,EmptyParam,EmptyParam,EmptyParam,EmptyParam,
         EmptyParam,EmptyParam,EmptyParam,EmptyParam,v,Emptyparam);
end;
  

function TARExcelProvider.LastUsedRow : integer;
begin
 result := ws.UsedRange.rows.Count;
end;

function TARExcelProvider.LastUsedColumn : integer;
begin
 result := ws.UsedRange.columns.Count;
end;

procedure TARExcelProvider.CleanUp; 
begin
  ea.ActiveWindow.LargeScroll(emptyparam,1000,emptyparam,emptyparam);
  ea.Calculation := xlCalculationAutomatic;
  ea.ScreenUpdating := true;
  ea.EnableEvents := true;
  ws.Calculate;
end;

procedure TARExcelProvider.AutoFit(col : integer); 
begin
 ws.Cells.Item[1,col].entirecolumn.autofit;
end; 

procedure TARExcelProvider.ActivateCell(row,col : integer); 
begin
  ws.Cells.Item[row,col].Activate;
end; 

procedure TARExcelProvider.ShowLevels(level : integer);
begin
 ws.Outline.ShowLevels(level,emptyparam);
end;

procedure TARExcelProvider.Run(macro : string); 
begin
 {$IFDEF FPC}
 ea.Run(widestring(macro));
 {$ELSE}
 ea.Run(macro);
 {$ENDIF}
end;

procedure TARExcelProvider.Print; 
begin
   ws.PrintOut(emptyparam,emptyparam,emptyparam,emptyparam,emptyparam, emptyparam,emptyparam,emptyparam);
end;

procedure TARExcelProvider.SaveDocAs(s:string); 
begin
 {$IFDEF FPC}
 wbk.SaveAs(widestring(s),xlNormal,widestring(''),widestring(''),false,false,xlNoChange,
           xlUserResolution,false,EmptyParam,EmptyParam);
 {$ELSE}
 wbk.SaveAs(s,xlNormal,'','',false,false,xlNoChange,
           xlUserResolution,false,EmptyParam,EmptyParam);
 {$ENDIF}
end;

procedure TARExcelProvider.SaveDoc; 
begin
  wbk.Save;
end;

procedure TARExcelProvider.BeforeQuit; 
begin
  ea.CutCopyMode := 0;
end;  

procedure TARExcelProvider.CloseDoc; 
begin
  wbk.Close(false,emptyparam,emptyparam);
end;

procedure TARExcelProvider.MarkSaved; 
begin
  wbk.Saved := true;
end;
           
procedure TARExcelProvider.Quit;
begin
  ws := Unassigned;
  ws := null;
  wbk := Unassigned;
  wbk := null;
  ea.quit;
  ea := Unassigned;
  ea := null;
end;

function TARExcelProvider.GetDate(val : TDateTime): OLEVariant;
begin
   Result := Val;
end;

function TARExcelProvider.GetMyVariant(Field: TField;str:boolean): OLEVariant;
begin
   Result := '';
  case Field.DataType of
    ftString, ftFixedChar, ftWideString, ftMemo, ftFmtMemo:
    if str then
     Result := '''' + Field.AsString
    else
     Result := Field.AsString;
    ftSmallint, ftInteger, ftWord, ftLargeint, ftAutoInc: Result := Field.AsInteger;
    ftFloat, ftCurrency, ftBCD: Result := Field.AsFloat;
    ftBoolean: Result := Field.AsBoolean;
    ftDate, ftTime, ftDateTime: Result := Field.AsDateTime;
  end;
end;

function TARExcelProvider.DocLoaded : boolean;
begin
 result := not (VarIsEmpty(wbk) or VarIsNull(wbk));
end;

procedure TARExcelProvider.Group(row, count : integer);
begin
 {$IFDEF FPC}
// ws.Range[widestring('A'+inttostr(row)),widestring('A'+inttostr(row+count - 1))].Rows.Group(emptyparam,emptyparam,emptyparam,emptyparam);
 ws.range[ws.rows.item[row],ws.rows.item[row+count-1]].Rows.Group(emptyparam,emptyparam,emptyparam,emptyparam);
 {$ELSE}
 ws.Range['A'+inttostr(row),'A'+inttostr(row+count - 1)].Rows.Group(emptyparam,emptyparam,emptyparam,emptyparam);
 {$ENDIF}
end;

end.
