unit artreport;

{$I artrep.inc}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface
uses classes,db,sysutils,artrepprovider;

type

  TArtRepException = class(Exception)
  end;

  TXPos = (beginpos, endpos, masterbpos, masterepos, detailbpos, detailepos,
           subdetailbpos, subdetailepos, groupbpos, groupepos,
           subgroupbpos, subgroupepos);
  TReportType = (rtnone,rtsimple,rtdetail,rtsubdetail,rtgroup,rtsubgroup);

  TArtRepValueProc = procedure(name:pchar;var Val:Variant) of object;
  TArtRepColumnsProc = procedure(group:integer;var Count:integer) of object;
  TArtRepProgressProc = procedure(num,total:integer) of object;

  TGrpData = record
   row, count: integer;
  end;


  TArtReport = class(TComponent)
  private
   ftype : TReportType;
   ftemplate : string;
   fvalueproc : TArtRepValueproc;
   fcolumnsproc : TArtRepColumnsproc;
   fprogressproc : TArtRepProgressProc;
   fpage : integer;
   fcolumns : integer;
   fsaveas : string;
   fmacro : string;
   ferror: string;
   freccount : integer;
   fgrouping, fsubgrouping : boolean;
   fshowempty : boolean;
   fshowemptysub : boolean;
   fshowreport : boolean;
   fmaster, fdetail, fsubdetail : TDataSource;
   fmasterfield, fdetailfield, fsubmasterfield,
   fsubdetailfield, fgroupfield, fsubgroupfield : string;
   fmastername, fdetailname, fsubname : string;
   farraysize : integer;
   Pvd : TArtRepProvider;
   autofit : array of boolean;
   ccnt : integer;
   masterno, dtlgrpno, subno,
   dtlgrptotalno, subtotalno, suballno,
   dtlgrpcnt, subcnt, subtotal, dtlgrptotal,
   mastercnt, suball : integer;
   pos : array [TXPos] of integer;
   owner : TComponent;
   fprint : boolean;
   fvisible : boolean;
   grpdata : array of TGrpData;
   grpcount : integer;
   fSheetName : string;

	procedure checkmulticolumn(var s:string);
	procedure settemplate;
	function GetValue(s : string) : Variant;
	function GetValueEx(s : string;level : integer = 1;str : boolean = true) : variant;
	procedure makesimple;
	procedure AddGrp(row,cnt:integer);
	procedure makedetail;
	procedure makesubdetail;
	procedure makegroup;
	procedure makesubgroup;
	procedure make;
	function IntReport:integer;

   procedure SetField(s:string;var f:string; var d:TDatasource);
   procedure SetMaster(s:string);
   procedure SetDetail(s:string);
   procedure SetSubDetail(s:string);
   procedure SetGroup(s:string);
   procedure SetSubgroup(s:string);
   procedure SetSubMaster(s:string);

  public
   constructor Create(parent:TComponent); override;
   destructor Destroy; override;
   procedure Report;

  published
   property Template : string read ftemplate write ftemplate;
   property Page : integer read fPage write fPage default 0;
   property SaveAs : string read fSaveas write fSaveas;
   property ShowReport : boolean read fShowreport write fshowreport default false;
   property MasterSource : TDataSource read fMaster write fMaster;
   property DetailSource : TDataSource read fDetail write fDetail;
   property SubDetailSource : TDataSource read fSubdetail write fsubdetail;
   property MasterField : string read fmasterfield write SetMaster;
   property DetailField : string read fDetailfield write SetDetail;
   property SubMasterField : string read fsubmasterfield write SetsubMaster;
   property SubDetailField : string read fSubDetailfield write SetSubDetail;
   property GroupField : string read fGroupfield write SetGroup;
   property SubGroupField : string read fsubGroupfield write SetsubGroup;
   property OnGetValue : TArtRepValueProc read fvalueproc write fvalueproc;
   property OnGetColumns : TArtRepColumnsProc read fcolumnsproc write fcolumnsproc;
   property OnProgress : TArtRepProgressProc read fprogressproc write fprogressproc;
   property ShowEmpty : boolean read fshowempty write fshowempty default false;
   property ShowEmptySub : boolean read fshowemptysub write fshowemptysub default false;
   property Grouping : boolean read fgrouping write fgrouping default false;
   property SubGrouping : boolean read fsubgrouping write fsubgrouping default false;
   property MasterRangeName : string read fmastername write fmastername;
   property RangeName : string read fdetailname write fdetailname;
   property SubRangeName : string read fsubname write fsubname;
   property Columns : integer read fcolumns write fcolumns default 0;
   property RecCount : integer read freccount write freccount;
   property RunMacro : string read fmacro write fmacro;
   property ArraySize : integer read farraysize write farraysize default 1000;
   property Print : boolean read fPrint write fPrint default false;
   property SheetName : string read fSheetName write fSheetName;
   property Error : string read ferror write ferror;
   property Provider : TArtRepProvider read Pvd write Pvd;
   property Visible : boolean read fvisible write fvisible default false;
  end;

  {$IFDEF ARTXLSREPORT}
  TArtXLSReport = class(TArtReport)
  end;
  {$ENDIF}


procedure CloseExcel;
procedure Register;

implementation


uses variants,oleserver,activex,shellapi,windows,comobj{$IFDEF FPC}, Fileutil,LResources{$ENDIF};
 {$DEFINE MyVariant}

constructor TArtReport.Create(parent:Tcomponent);
begin
 inherited Create(parent);

 ftype := rtnone;
 fmastername := 'Base';
 fdetailname := 'Base';
 fsubname := 'Base';
 farraysize := 1000;
end;

destructor TArtReport.Destroy;
begin
 inherited Destroy;
end;

procedure TArtReport.SetField(s:string;var f:string; var d:TDatasource);
var L:TStringList;
begin
 if (s <> f) and
    (d <> nil) and (d.DataSet <> nil) then
 begin
   L := TStringList.Create;
   L.CaseSensitive := false;
   d.DataSet.GetFieldNames(L);
   if L.IndexOf(s) <> -1 then f := s;
   L.Free;
   L := nil;
 end;
end;


procedure TArtReport.SetMaster(s:string);
begin
 SetField(s,fmasterfield,fMaster);
end;

procedure TArtReport.SetSubMaster(s:string);
begin
 SetField(s,fsubmasterfield,fDetail);
end;

procedure TArtReport.SetDetail(s:string);
begin
 SetField(s,fdetailfield,fDetail);
end;

procedure TArtReport.SetSubDetail(s:string);
begin
 SetField(s,fsubdetailfield,fsubdetail);
end;

procedure TArtReport.SetGroup(s:string);
begin
 SetField(s,fgroupfield,fMaster);
end;

procedure TArtReport.Setsubgroup(s:string);
begin
 SetField(s,fsubgroupfield,fMaster);
end;

procedure CloseExcel;
var
   EA : Variant;
begin
   try
    EA := GetActiveOleObject('Excel.Application');
    EA.Quit;
    ea := Unassigned;
    ea := null;
  except
  end;
end;

procedure TArtReport.Report;
begin
  ferror := '';
  if IntReport <> 0 then raise TArtRepException.Create(ferror);
end;

procedure Register;
begin
 {$IFDEF ARTXLSREPORT}
   RegisterComponents('Art', [TArtReport,TArtXLSReport]);
 {$ELSE}  
   RegisterComponents('Art', [TArtReport]);
 {$ENDIF}  
end;

{$IFDEF Myvariant}
function xGetMyVariant(Field: TField;str:boolean = true): OLEVariant;
begin
   Result := '';
  case Field.DataType of
    ftString, ftFixedChar, ftWideString, ftMemo, ftFmtMemo:
    if str then
     Result := '''' + Field.AsString
    else
     Result := Field.AsString;
    ftSmallint, ftInteger, ftWord, ftLargeint, ftAutoInc: Result := Field.AsInteger;
    ftFloat, ftCurrency, ftBCD: Result := Field.AsFloat;
    ftBoolean: Result := Field.AsBoolean;
    ftDate, ftTime, ftDateTime: Result := Field.AsDateTime;
  end;
end;
{$EndIf}

procedure split(s:string;var ds : char; var nm:string);
var i:integer; dsnm : string;
begin
  i := pos('.',s);
  ds := ' ';
  nm := s;
  if i <> 0 then
  begin
   dsnm := Uppercase(copy(s,1,i-1));
   nm := copy(s,i+1,length(s)-i);
   if (dsnm = 'MASTER') or (dsnm = 'M') then ds := 'M'
   else
   if (dsnm = 'DETAIL') or (dsnm = 'D') then ds := 'D'
   else
   if (dsnm = 'SUBDETAIL') or (dsnm = 'S') or (dsnm = 'SUB') then ds := 'S'
   else
    ds := '!';
  end;
end;

procedure TArtReport.checkmulticolumn(var s:string);
var bmult,emult,i,j,k,ccnt,rcnt,bpos,epos,grp,clmn:integer;
    x : string;
    //buf : array[0..255] of char;
    first : boolean;
begin
 {$IFDEF FPC}
  s := ExpandFileNameUTF8(s); { *Converted from ExpandFileName*  }
 {$ELSE}
  s := ExpandFileName(s);
 {$ENDIF}
  if not {$IFDEF FPC} FileExistsUTF8(s) { *Converted from FileExists*  }
         {$ELSE} FileExists(s) {$ENDIF}  then raise TArtRepException.Create('Template not found!');
 try
  Pvd.Connect;
  Pvd.Setvisible(fvisible);

  if fpage = 0 then
  begin
   Pvd.CreateDoc(ftemplate);
  end
  else
  begin
   Pvd.OpenDoc(ftemplate,fpage);
  end;

  Pvd.SetCalculation(fvisible);
  
  x := UpperCase(Pvd.GetCelltext(1,1));
  if x = 'MULTICOLUMN' then
  begin
   Pvd.DeleteRow(1);

   first := true;
   grp := 1;

  repeat
   bmult := 0;
   emult := 0;
   ccnt := Pvd.LastUsedColumn;
   rcnt := Pvd.LastUsedRow;
   for i := 1 to ccnt do
   begin
    if (bmult <> 0) and (emult <> 0) then break;
    x := UpperCase(Pvd.GetCelltext(1,i));
    if x = '#BEGIN' then
      bmult := i
    else
    if x = '#END' then
      emult := i-2;
   end;
   if first and ((emult < bmult) or (emult = 0) or (bmult = 0)) then
    raise TArtRepException.Create('Wrong Template!');

   first := false;

   if emult = 0 then break;

   bpos := 0;
   epos := 0;

   for i := 4 to rcnt do
   begin
    if (bpos <> 0) and (epos <> 0) then break;
    x := UpperCase(Pvd.GetCelltext(i,1));
    if x = '#BEGIN' then
      bpos := i-2
    else
    if x = '#END' then
      epos := i-2;
   end;
   if (epos < bpos) or (epos = 0) or (bpos = 0) then
    raise TArtRepException.Create('Wrong Template!');

   Pvd.DeleteRow(1);
   Pvd.DeleteColumn(emult+2);
   Pvd.DeleteColumn(bmult);


   clmn := fcolumns;
   if Assigned(fcolumnsproc) then fcolumnsproc(grp,clmn);
   for i := 1 to clmn-1 do
   begin
   
    Pvd.InsertColumns(emult+1,emult-bmult+1);
    Pvd.CopyColumns(bmult,emult+1,emult-bmult+1); 
        
    for j := bpos + 1 to epos - 1 do
     for k := emult+1 to emult+emult-bmult+1 do
     begin
      x := UpperCase(Pvd.GetCelltext(j,k));
      if (length(x) > 2) and (copy(x,1,2)='$$') then
       Pvd.SetCellText(j,k,copy(x,2,length(s)-1)+inttostr(clmn-i+1))
      else
      if (length(x) > 2) and (copy(x,1,2)='&&') then
       Pvd.SetCellText(j,k,copy(x,2,length(s)-1)+inttostr(clmn-i+1));
     end;
   end;
   for j := bpos + 1 to epos - 1 do
     for k := bmult to emult do
     begin
      x := UpperCase(Pvd.GetCelltext(j,k));
      if (length(x) > 2) and (copy(x,1,2)='$$') then
       Pvd.SetCellText(j,k,copy(x,2,length(s)-1)+'1')
      else
      if (length(x) > 2) and (copy(x,1,2)='&&') then
       Pvd.SetCellText(j,k,copy(x,2,length(s)-1)+'1');
     end;
   inc(grp);
  until bmult = 0;


  end;

 finally
 end;
end;

procedure TArtReport.settemplate;
var x : string;
begin
 try
  x := UpperCase(Pvd.GetCellText(1,1));
  if x = 'SIMPLE' then
   ftype := rtsimple
  else
  if x = 'DETAIL' then
   ftype := rtdetail
  else
  if x = 'SUBDETAIL' then
   ftype := rtsubdetail
  else
  if x = 'GROUP' then
   ftype := rtgroup
  else
  if x = 'SUBGROUP' then
   ftype := rtsubgroup
  else
   raise TArtRepException.Create('Wrong template!');
 finally
 end;
end;


function TArtReport.GetValue(s : string) : Variant;
var nm,s1,s2:string; x,y:integer;
begin
    if (length(s) = 0) or (s[1]<>'&') then
    begin
     result := s;
     exit;
    end;
    result := '';
    s := copy(s,2,length(s)-1);
    s1 := uppercase(s);
    if (s1 = 'DATE') then result := Pvd.GetDate(Now)
    else
    if (s1 = 'MASTERCOUNT') or (s1 = 'COUNT') then
     result := 'MasterCount'
    else
    if (s1 = 'DETAILTOTAL') then
    begin
     if (ftype = rtdetail) or (ftype = rtsubdetail) then
      result := 'DetailTotal'
     else raise TArtRepException.Create('Error in cell: DETAILTOTAL in not detail report!');
    end
    else
    if (s1 = 'SUBDETAILALL') then
    begin
     if ftype = rtsubdetail then
      result := 'SubDetailAll'
     else raise TArtRepException.Create('Error in cell: SUBDETAILALL in not sub-detail report!');
    end
    else
    if (s1 = 'GROUPTOTAL') then
    begin
     if (ftype = rtgroup) or (ftype = rtsubgroup) then
      result := 'GroupTotal'
     else raise TArtRepException.Create('Error in cell: GROUPTOTAL in not grouped report!');
    end
    else
    if (s1 = 'SUBGROUPALL') then
    begin
     if ftype = rtsubgroup then
      result := 'SubGroupAll'
     else raise TArtRepException.Create('Error in cell: SUBGROUPALL in not sub-grouped report!');
    end
    else
    if (s[1] = '=') or (s[1]='?') then
    begin
     if s[1] = '?' then
         s := copy(s,2,length(s)-1);
     x := system.Pos('{',s);
     while x > 0 do
     begin
      y := system.Pos('}',s);
      if y < x then raise TArtRepException.Create('Error in cell: brace balance report!');
      nm := copy(s,x+1,y-x-1);
      s1 := copy(s,1,x-1);
      s2 := copy(s,y+1,length(s)-y);
      nm := GetValue(nm);
      s := s1 + nm + s2;
      x := system.Pos('{',s);
     end;
      result := s;
    end
    else
     if Assigned(fvalueproc) then
      fvalueproc(pchar(uppercase(s)),result);
end;

function TArtReport.GetValueEx(s : string;level : integer = 1;str : boolean = true) : variant;
var nm,s1,s2:string; x,y:integer; ds:char;
begin
    if (length(s) = 0) or ((s[1]<>'&') and (s[1]<>'$')) then
    begin
     result := s;
     exit;
    end;
    result := '';
    if s[1] = '$' then
    begin
     nm := uppercase(copy(s,2,length(s)-1));
     split(nm,ds,s);
       case ds of
{$ifdef myvariant}
       'M': result := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s),str);
       'D': if level > 1 then
             result := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s),str)
            else raise TArtRepException.Create('Wrong template!');
       'S': if level > 2 then
             result := Pvd.GetMyVariant(fsubdetail.dataset.FieldbyName(s),str)
            else raise TArtRepException.Create('Wrong template!');
       ' ': case level of
            1: result := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s),str);
            2: result := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s),str);
            3: result := Pvd.GetMyVariant(fsubdetail.dataset.FieldbyName(s),str);
            end;
{$else}
       'M': result := fmaster.dataset.FieldbyName(s).Value; //GetMyVariant(fmaster.dataset.FieldbyName(nm));
       'D': if level > 1 then
             result := fdetail.dataset.FieldbyName(s).Value //GetMyVariant(fdetail.dataset.FieldbyName(nm));
            else raise TArtRepException.Create('Wrong template!');
        'S':if level > 2 then
             result := fsubdetail.dataset.FieldbyName(s).Value//  GetMyVariant(fsubdetail.dataset.FieldbyName(nm));
            else raise TArtRepException.Create('Wrong template!');
       ' ': case level of
            1: result := fmaster.dataset.FieldbyName(s).Value;
            2: result := fdetail.dataset.FieldbyName(s).Value;
            3: result := fsubdetail.dataset.FieldbyName(s).Value;
            end;
{$endif}
       else raise TArtRepException.Create('Wrong template!');
       end;
     exit;
    end;
    s := copy(s,2,length(s)-1);
    s1 := uppercase(s);
    if (s1 = 'MASTERCOUNT') or (s1 = 'COUNT') then
      result := 'MasterCount'
    else
    if (s1 = 'DETAILCOUNT') then
    begin
     if (ftype = rtdetail) or (ftype = rtsubdetail) then
      result := 'DetailCount'+inttostr(masterno)
     else raise TArtRepException.Create('Error in datacell: DETAILCOUNT in not detail report!');
    end
    else
    if (s1 = 'SUBDETAILCOUNT') then
    begin
     if ftype = rtsubdetail then
      result := 'SubDetailCount'+inttostr(masterno)+'_'+inttostr(dtlgrpno)
     else raise TArtRepException.Create('Error in datacell: SUBDETAILCOUNT in not sub-detail report!');
    end
    else
    if (s1 = 'DETAILTOTAL') then
    begin
     if (ftype = rtdetail) or (ftype = rtsubdetail) then
      result := 'DetailTotal'
     else raise TArtRepException.Create('Error in datacell: DETAILTOTAL in not detail report!');
    end
    else
    if (s1 = 'SUBDETAILTOTAL') then
    begin
     if ftype = rtsubdetail then
      result := 'SubDetailTotal'+inttostr(masterno)
     else raise TArtRepException.Create('Error in datacell: SUBDETAILTOTAL in not sub-detail report!');
    end
    else
    if (s1 = 'SUBDETAILALL') then
    begin
     if ftype = rtsubdetail then
      result := 'SubDetailAll'
     else raise TArtRepException.Create('Error in datacell: SUBDETAILALL in not sub-detail report!');
    end
    else
    if (s1 = 'GROUPCOUNT') then
    begin
     if (ftype = rtgroup) or (ftype = rtsubgroup) then
      result := 'GroupCount'+inttostr(masterno)
     else raise TArtRepException.Create('Error in datacell: GROUPCOUNT in not grouped report!');
    end
    else
    if (s1 = 'SUBGROUPCOUNT') then
    begin
     if ftype = rtsubgroup then
      result := 'SubGroupCount'+inttostr(masterno)+'_'+inttostr(dtlgrpno)
     else raise TArtRepException.Create('Error in datacell: SUBGROUPCOUNT in not sub-grouped report!');
    end
    else
    if (s1 = 'SUBGROUPTOTAL') then
    begin
     if ftype = rtsubgroup then
      result := 'SubGroupTotal'+inttostr(masterno)
     else raise TArtRepException.Create('Error in datacell: SUBGROUPTOTAL in not sub-grouped report!');
    end
    else
    if (s1 = 'GROUPTOTAL') then
    begin
     if (ftype = rtgroup) or (ftype = rtsubgroup) then
      result := 'GroupTotal'
     else raise TArtRepException.Create('Error in datacell: GROUPTOTAL in not grouped report!');
    end
    else
    if (s1 = 'SUBGROUPALL') then
    begin
     if ftype = rtsubgroup then
      result := 'SubGroupAll'
     else raise TArtRepException.Create('Error in datacell: SUBGROUPALL in not sub-grouped report!');
    end
    else

    if (s1 = 'RECNO') or (s1 = 'MASTERRECNO')  then
      result := masterno
    else
    if s1 = 'DETAILRECNO' then
    begin
      if (ftype = rtdetail) or (ftype = rtsubdetail) then
        result := dtlgrpno
      else raise TArtRepException.Create('Error in datacell: DETAILRECNO in not detail report!');
    end
    else
    if s1 = 'GROUPRECNO' then
    begin
      if (ftype = rtgroup) or (ftype = rtsubgroup) then
        result := dtlgrpno
      else raise TArtRepException.Create('Error in datacell: GROUPRECNO in not grouped report!');
    end
    else
    if s1 = 'SUBDETAILRECNO' then
    begin
      if ftype = rtsubdetail then
        result := subno
      else raise TArtRepException.Create('Error in datacell: SUBDETAILRECNO in not sub-detail report!');
    end
    else
    if s1 = 'SUBGROUPRECNO' then
    begin
      if ftype = rtsubgroup then
        result := subno
      else raise TArtRepException.Create('Error in datacell: SUBGROUPRECNO in not sub-grouped report!');
    end
    else
    if s1 = 'DETAILTOTALRECNO' then
    begin
      if (ftype = rtdetail) or (ftype = rtsubdetail) then
        result := dtlgrptotalno
      else raise TArtRepException.Create('Error in datacell: DETAILTOTALRECNO in not detail report!');
    end
    else
    if s1 = 'SUBDETAILTOTALRECNO' then
    begin
      if ftype = rtsubdetail then
        result := subtotalno
      else raise TArtRepException.Create('Error in datacell: SUBDETAILTOTALRECNO in not sub-detail report!');
    end
    else
    if s1 = 'SUBDETAILALLRECNO' then
    begin
      if ftype = rtsubdetail then
        result := suballno
      else raise TArtRepException.Create('Error in datacell: SUBDETAILALLRECNO in not sub-detail report!');
    end
    else
    if s1 = 'GROUPTOTALRECNO' then
    begin
      if (ftype = rtgroup) or (ftype = rtsubgroup) then
        result := dtlgrptotalno
      else raise TArtRepException.Create('Error in datacell: GROUPTOTALRECNO in not grouped report!');
    end
    else
    if s1 = 'SUBGROUPTOTALRECNO' then
    begin
      if ftype = rtsubgroup then
        result := subtotalno
      else raise TArtRepException.Create('Error in datacell: SUBGROUPTOTALRECNO in not sub-grouped report!');
    end
    else
    if s1 = 'SUBGROUPALLRECNO' then
    begin
      if ftype = rtsubgroup then
        result := suballno
      else raise TArtRepException.Create('Error in datacell: SUBGROUPALLRECNO in not sub-grouped report!');
    end
    else
    if (s1 = 'DATE') then result := Pvd.GetDate(Now)
    else
    if (s[1] = '=') or (s[1] = '?') then
    begin
     if s[1] = '?' then
         s := copy(s,2,length(s)-1);
     x := system.Pos('{',s);
     while x > 0 do
     begin
      y := system.Pos('}',s);
      if y < x then raise TArtRepException.Create('Error in datacell: braces balance!');
      nm := copy(s,x+1,y-x-1);
      s1 := copy(s,1,x-1);
      s2 := copy(s,y+1,length(s)-y);
      nm := GetValueEx(nm,level,false);
      s := s1 + nm + s2;
      x := system.Pos('{',s);
     end;
      result := s;
    end
    else
     if Assigned(fvalueproc) then fvalueproc(pchar(uppercase(s)),result);
end;


procedure TArtReport.makesimple;
var dpos, dlen, i,j,k,k1,xp,rcnt:integer;
    fields : array of array of string;
    nm : string;
    ArrayData  : Variant;
begin
  dlen := pos[masterepos] - pos[masterbpos]+1;
  dpos := pos[masterbpos];
  setlength(fields,dlen,ccnt);

  for i := 0 to dlen-1 do
   for j := 1 to ccnt do
    fIELDS[i,j-1] := Pvd.GetCellFormula(dpos+i,j);

  i := 1;
  if freccount <= 0 then
   rcnt := fmaster.dataset.RecordCount
  else 
   rcnt := freccount;
  k := rcnt-1; 
  mastercnt := rcnt;
  while i < rcnt do
  begin
   if k >= i then k1 := i else k1 := k;
   
   Pvd.InsertRows(dpos+dlen*k1,dlen*k1);
   Pvd.CopyRows(dpos,dpos+dlen*k1,dlen*k1);
   inc(i,k1);
   dec(k,k1);
  end;                                

  ArrayData := VarArrayCreate([1, farraysize*dlen, 1, ccnt], varVariant);
  fmaster.dataset.First;

  k1 := 0;
  xp := 0;
  fmaster.DataSet.First;
  for i := 1 to rcnt do
  begin
   masterno := i;
   if assigned(fprogressproc) then
     fprogressproc(i,rcnt);
   
   for j := 1 to dlen do
   begin
   inc(xp);

   for k := 1 to ccnt do
   begin
    ArrayData[xp,k] := Fields[j-1][k-1];
    if fields[j-1][k-1] <> '' then
    begin
     if fields[j-1][k-1][1] = '&' then
     begin
      ArrayData[xp,k] := getvalueex(fields[j-1][k-1]);
     end
     else
     if fields[j-1][k-1][1] = '$' then
     begin
      nm := uppercase(copy(fields[j-1][k-1],2,length(fields[j-1][k-1])-1));
{$IFDEF MyVariant}
      ArrayData[xp,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$Else}
      ArrayData[xp,k] := fmaster.dataset.FieldbyName(nm).Value;//GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$EndIF}
     end;
    end;
   end;
   end;
   fmaster.dataset.Next;
   inc(k1);
   if k1 mod farraysize = 0 then
   begin

    Pvd.SetRange(dpos+(k1-farraysize)*dlen,1,dpos+k1*dlen - 1,ccnt,ArrayData);
    xp := 0;
   end;
  end;

  if k1 mod farraysize <> 0 then
  begin
    Pvd.SetRange(dpos+(k1-k1 mod farraysize)*dlen,1,dpos+k1*dlen - 1,ccnt,ArrayData);
  end;

  Pvd.AddNameRange(dpos,1,dpos+rcnt*dlen - 1,1,fmastername);
  Pvd.AddName('MasterCount',mastercnt);

  ArrayData := unassigned;
end;


procedure TArtReport.AddGrp(row,cnt:integer);
begin
 if cnt = 0 then exit;
 if grpcount > High(grpdata) then
  setlength(grpdata,grpcount+50);
 grpdata[grpcount].row := row;
 grpdata[grpcount].count := cnt;
 inc(grpcount);
end;

procedure TArtReport.makedetail;
var dpos, mtpos, mbpos, allen, tkpos,
    mtlen, mblen, dlen, i,j,k,i2,k1,k2,
    dcnt, rcnt:integer;
    mtfields, dfields, mbfields : array of array of string;
    s,nm : string;
    ArrayData  : Variant;
    doit : boolean;
    ds : char;
begin
  grpcount := 0;
  dlen := pos[detailepos] - pos[detailbpos]+1;
  mtlen := pos[detailbpos] - pos[masterbpos];
  mblen := pos[masterepos] - pos[detailepos];
  allen := pos[masterepos] - pos[masterbpos]+1;
  dpos := pos[detailbpos];
  mtpos := pos[masterbpos];
  mbpos := pos[detailepos]+1;
  setlength(dfields,dlen,ccnt);
  if mtlen > 0 then setlength(mtfields,mtlen,ccnt);
  if mblen > 0 then setlength(mbfields,mblen,ccnt);

  for i := 0 to dlen-1 do
   for j := 1 to ccnt do
    dfIELDS[i,j-1] := Pvd.GetCellFormula(dpos+i,j);

  for i := 0 to mtlen-1 do
   for j := 1 to ccnt do
    mtfIELDS[i,j-1] := Pvd.GetCellFormula(mtpos+i,j);

  for i := 0 to mblen-1 do
   for j := 1 to ccnt do
    mbfIELDS[i,j-1] := Pvd.GetCellFormula(mbpos+i,j);

  ArrayData := VarArrayCreate([1, farraysize*dlen, 1, ccnt], varVariant);
  fmaster.dataset.First;
  tkpos := pos[masterepos]+1;

  dtlgrptotal := 0;
  dtlgrptotalno := 0;
  masterno := 0;
  mastercnt := 0;
  fmaster.DataSet.First;

  rcnt := fmaster.dataset.RecordCount;
  for i := 1 to rcnt do
  begin

    if assigned(fprogressproc) then
     fprogressproc(i,rcnt);


   dtlgrpcnt := 0;
   dtlgrpno := 0;
   dcnt := 0;
   doit := fdetail.DataSet.Locate(fdetailfield,
      fmaster.dataset.fieldbyname(fmasterfield).AsVariant,[]);
   if doit then
   while (fdetail.DataSet.fieldbyname(fdetailfield).asvariant =
         fmaster.dataset.fieldbyname(fmasterfield).AsVariant)
         and not fdetail.dataset.Eof do
   begin
    fdetail.DataSet.Next;
    inc(dcnt);
   end;
   if not doit and fshowempty then doit := true;
   if doit then
   begin
    inc(masterno);
    inc(mastercnt);

    Pvd.InsertRows(tkpos,allen);
    Pvd.CopyRows(mtpos,tkpos,allen);
    for j := 1 to mtlen do
    for k := 1 to ccnt do
    begin
     ArrayData[j,k] := mtFields[j-1][k-1];
     if mtfields[j-1][k-1] <> '' then
     begin
      if mtfields[j-1][k-1][1] = '&' then
      begin
       ArrayData[j,k] := getvalueex(mtfields[j-1][k-1]);
      end
      else
      if mtfields[j-1][k-1][1] = '$' then
      begin
       nm := uppercase(copy(mtfields[j-1][k-1],2,length(mtfields[j-1][k-1])-1));
       split(nm,ds,s);
       case ds of
{$IFDEF MyVariant}
       'M',' ': ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s));
       'D': ArrayData[j,k] := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s));
{$Else}
       'M',' ': ArrayData[j,k] := fmaster.dataset.FieldbyName(s).Value;//GetMyVariant(fmaster.dataset.FieldbyName(nm));
       'D': ArrayData[j,k] := fdetail.dataset.FieldbyName(s).VALUE;// GetMyVariant(fdetail.dataset.FieldbyName(nm));
{$Endif}
       else raise TArtRepException.Create('Wrong template!');
       end;
      end;
     end;
    end;
    if mtlen > 0 then Pvd.SetRange(tkpos,1,tkpos+mtlen - 1,
          ccnt, ArrayData);


    for j := 1 to mblen do
    for k := 1 to ccnt do
    begin
     ArrayData[j,k] := mbFields[j-1][k-1];
     if mbfields[j-1][k-1] <> '' then
     begin
      if mbfields[j-1][k-1][1] = '&' then
      begin
       ArrayData[j,k] := getvalueex(mbfields[j-1][k-1]);
      end
      else
      if mbfields[j-1][k-1][1] = '$' then
      begin
       nm := uppercase(copy(mbfields[j-1][k-1],2,length(mbfields[j-1][k-1])-1));
       split(nm,ds,s);
       case ds of
{$IFDef MyVariant}
       'M',' ': ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s));
       'D': ArrayData[j,k] := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s));
{$else}
       'M',' ': ArrayData[j,k] := fmaster.dataset.FieldbyName(s).Value;//  GetMyVariant(fmaster.dataset.FieldbyName(nm));
       'D': ArrayData[j,k] := fdetail.dataset.FieldbyName(s).Value;//  GetMyVariant(fdetail.dataset.FieldbyName(nm));
{$endif}
       else raise TArtRepException.Create('Wrong template!');
       end;
      end;
     end;
    end;
    if mblen > 0 then Pvd.SetRange(tkpos+mbpos-mtpos,1,tkpos+mbpos-mtpos+mblen-1,
                ccnt, ArrayData);

    inc(tkpos,mtlen);

    if dcnt = 0 then
     Pvd.DeleteRows(tkpos,dlen)
    else
    begin
     fdetail.DataSet.Locate(fdetailfield,
         fmaster.dataset.fieldbyname(fmasterfield).AsVariant,[]);

     i2 := 1;
     k := dcnt-1;
     while i2 < dcnt do
     begin
      if k >= i2 then k1 := i2 else k1 := k;

    Pvd.InsertRows(tkpos+dlen*k1,dlen*k1);
    Pvd.CopyRows(tkpos,tkpos+dlen*k1,dlen*k1);

      inc(i2,k1);
      dec(k,k1);
     end;

     inc(dtlgrptotal,dcnt);
     k1 := 0;
     k2 := 0;
     for i2 := 1 to dcnt do
     begin
      dtlgrpno := i2;
      inc(dtlgrpcnt);
      inc(dtlgrptotalno);
      for j := 1 to dlen do
      for k := 1 to ccnt do
      begin
       ArrayData[k2+j,k] := dFields[j-1][k-1];
       if dfields[j-1][k-1] <> '' then
       begin
        if dfields[j-1][k-1][1] = '&' then
        begin
         ArrayData[k2+j,k] := getvalueex(dfields[j-1][k-1],2);
        end
        else
        if dfields[j-1][k-1][1] = '$' then
        begin
         nm := uppercase(copy(dfields[j-1][k-1],2,length(dfields[j-1][k-1])-1));
         split(nm,ds,s);
         case ds of
{$Ifdef myvariant}
         'M': ArrayData[k2+j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s));
         'D',' ': ArrayData[k2+j,k] := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s));
{$else}
         'M': ArrayData[k2+j,k] := fmaster.dataset.FieldbyName(s).Value;//  GetMyVariant(fmaster.dataset.FieldbyName(nm));
         'D',' ': ArrayData[k2+j,k] := fdetail.dataset.FieldbyName(s).Value;//  GetMyVariant(fdetail.dataset.FieldbyName(nm));
{$endif}
         else raise TArtRepException.Create('Wrong template!');
         end;
        end;
       end;
      end;
      inc(k2,dlen);
      fdetail.dataset.Next;
      inc(k1);
      if k1 mod farraysize = 0 then
      begin
       Pvd.SetRange(tkpos+(k1-farraysize)*dlen,1,tkpos+k1*dlen - 1,
          ccnt, ArrayData);
       k2 := 0;
      end;
     end;

     if k1 mod farraysize <> 0 then
     begin
       Pvd.SetRange(tkpos+(k1-k1 mod farraysize)*dlen,1,tkpos+k1*dlen - 1,
        ccnt, ArrayData);
     end;

    end;



    if fgrouping then
    begin
     AddGrp(tkpos,dtlgrpcnt*dlen);
//     Pvd.Group(tkpos,dtlgrpcnt*dlen);
    end;


    Pvd.AddNameRange(tkpos,1,tkpos+dtlgrpcnt*dlen - 1,1,fdetailname+inttostr(masterno));
    Pvd.AddName('DetailCount'+inttostr(masterno),dtlgrpcnt);

    inc(tkpos,dcnt*dlen);
    inc(tkpos,mblen);
   end;
   fmaster.DataSet.Next;
  end;

  Pvd.AddNameRange(mtpos,1,tkpos - 1,1,fmastername);

  for j:=0 to grpcount-1 do Pvd.Group(grpdata[j].row,grpdata[j].count);
  Pvd.DeleteRows(mtpos,allen);

  Pvd.AddName('MasterCount',mastercnt);
  Pvd.AddName('GroupTotal',dtlgrptotal);


  ArrayData := unassigned;
end;

procedure TArtReport.makesubdetail;
var dtpos, dbpos, spos, mtpos, mbpos, allen, tkpos,
    mtlen, mblen, dtlen, dblen, slen, i,j,k,i2,k1,k2,id,
    allend, xpos, dttkpos, dcnt, scnt, mttkpos, rcnt:integer;
    mtfields, dtfields, mbfields, dbfields, sfields : array of array of string;
    s,nm : string;
    ArrayData  : Variant;
    doit,doitd : boolean;
    ds : char;
begin
  grpcount := 0;
  mtlen := pos[detailbpos] - pos[masterbpos];
  mblen := pos[masterepos] - pos[detailepos];
  dtlen := pos[subdetailbpos] - pos[detailbpos];
  dblen := pos[detailepos] - pos[subdetailepos];
  slen := pos[subdetailepos] - pos[subdetailbpos] + 1;
  allen := pos[masterepos] - pos[masterbpos]+1;
  allend := pos[detailepos] - pos[detailbpos]+1;
  spos := pos[subdetailbpos];
  mtpos := pos[masterbpos];
  mbpos := pos[detailepos]+1;
  dtpos := pos[detailbpos];
  dbpos := pos[subdetailepos]+1;

  setlength(sfields,slen,ccnt);
  if mtlen > 0 then setlength(mtfields,mtlen,ccnt);
  if mblen > 0 then setlength(mbfields,mblen,ccnt);
  if dtlen > 0 then setlength(dtfields,dtlen,ccnt);
  if dblen > 0 then setlength(dbfields,dblen,ccnt);

  for i := 0 to slen-1 do
   for j := 1 to ccnt do
    sfIELDS[i,j-1] := Pvd.GetCellFormula(spos+i,j);

  for i := 0 to mtlen-1 do
   for j := 1 to ccnt do
    mtfIELDS[i,j-1] := Pvd.GetCellFormula(mtpos+i,j);

  for i := 0 to mblen-1 do
   for j := 1 to ccnt do
    mbfIELDS[i,j-1] := Pvd.GetCellFormula(mbpos+i,j);

  for i := 0 to dtlen-1 do
   for j := 1 to ccnt do
    dtfIELDS[i,j-1] := Pvd.GetCellFormula(dtpos+i,j);

  for i := 0 to dblen-1 do
   for j := 1 to ccnt do
    dbfIELDS[i,j-1] := Pvd.GetCellFormula(dbpos+i,j);

  ArrayData := VarArrayCreate([1,farraysize *slen, 1, ccnt], varVariant);

  fmaster.dataset.First;
  tkpos := pos[masterepos]+1;

  dtlgrptotal := 0;
  dtlgrptotalno := 0;
  masterno := 0;
  mastercnt := 0;
  suball := 0;
  suballno := 0;

// Master

  rcnt := fmaster.dataset.RecordCount;

  fmaster.DataSet.First;
  for i := 1 to rcnt do
  begin

    if assigned(fprogressproc) then
     fprogressproc(i,rcnt);


   dtlgrpcnt := 0;
   dcnt := 0;
   dtlgrpno := 0;
   subtotal := 0;
   subtotalno := 0;

   doit := fdetail.DataSet.Locate(fdetailfield,
      fmaster.dataset.fieldbyname(fmasterfield).AsVariant,[]);
   if doit then
   while (fdetail.DataSet.fieldbyname(fdetailfield).asvariant =
         fmaster.dataset.fieldbyname(fmasterfield).AsVariant)
         and not fdetail.dataset.Eof do
   begin
    fdetail.DataSet.Next;
    inc(dcnt);
   end;

   if not doit and fshowempty then doit := true;
   if doit then
   begin
    inc(masterno);
    inc(mastercnt);

    Pvd.InsertRows(tkpos,allen);
    Pvd.CopyRows(mtpos,tkpos,allen);
    for j := 1 to mtlen do
    for k := 1 to ccnt do
    begin
     ArrayData[j,k] := mtFields[j-1][k-1];
     if mtfields[j-1][k-1] <> '' then
     begin
      if mtfields[j-1][k-1][1] = '&' then
      begin
       ArrayData[j,k] := getvalueex(mtfields[j-1][k-1]);
      end
      else
      if mtfields[j-1][k-1][1] = '$' then
      begin
       nm := uppercase(copy(mtfields[j-1][k-1],2,length(mtfields[j-1][k-1])-1));
       split(nm,ds,s);
       case ds of
{$ifdef myvariant}
       'M',' ': ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s));
       'D': ArrayData[j,k] := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s));
       'S': ArrayData[j,k] := Pvd.GetMyVariant(fsubdetail.dataset.FieldbyName(s));
{$else}
       'M',' ': ArrayData[j,k] := fmaster.dataset.FieldbyName(s).Value; //GetMyVariant(fmaster.dataset.FieldbyName(nm));
       'D': ArrayData[j,k] := fdetail.dataset.FieldbyName(s).Value; //GetMyVariant(fdetail.dataset.FieldbyName(nm));
       'S': ArrayData[j,k] := fsubdetail.dataset.FieldbyName(s).Value;//  GetMyVariant(fsubdetail.dataset.FieldbyName(nm));
{$endif}
       else raise TArtRepException.Create('Wrong template!');
       end;
      end;
     end;
    end;

    if mtlen > 0 then Pvd.SetRange(tkpos,1,tkpos+mtlen - 1,
       ccnt, ArrayData);



    for j := 1 to mblen do
    for k := 1 to ccnt do
    begin
     ArrayData[j,k] := mbFields[j-1][k-1];
     if mbfields[j-1][k-1] <> '' then
     begin
      if mbfields[j-1][k-1][1] = '&' then
      begin
       ArrayData[j,k] := getvalueex(mbfields[j-1][k-1]);
      end
      else
      if mbfields[j-1][k-1][1] = '$' then
      begin
       nm := uppercase(copy(mbfields[j-1][k-1],2,length(mbfields[j-1][k-1])-1));
       split(nm,ds,s);
       case ds of
{$ifdef myvariant}
       'M',' ': ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s));
       'D': ArrayData[j,k] := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s));
       'S': ArrayData[j,k] := Pvd.GetMyVariant(fsubdetail.dataset.FieldbyName(s));
{$else}
       'M',' ': ArrayData[j,k] := fmaster.dataset.FieldbyName(s).Value;//  GetMyVariant(fmaster.dataset.FieldbyName(nm));
       'D': ArrayData[j,k] := fdetail.dataset.FieldbyName(s).Value; // GetMyVariant(fdetail.dataset.FieldbyName(nm));
       'S': ArrayData[j,k] := fsubdetail.dataset.FieldbyName(s).Value; // GetMyVariant(fsubdetail.dataset.FieldbyName(nm));
{$endif}
       else raise TArtRepException.Create('Wrong template!');
       end;
      end;
     end;
    end;
    if mblen > 0 then Pvd.SetRange(tkpos+mbpos-mtpos,1,tkpos+mbpos-mtpos+mblen-1,
       ccnt, ArrayData);


    inc(tkpos,mtlen);
    mttkpos := tkpos;
// Detail
    if dcnt = 0 then
     Pvd.DeleteRows(tkpos,allend)
    else
    begin
      Pvd.DeleteRows(tkpos,allend);

     fdetail.DataSet.Locate(fdetailfield,
         fmaster.dataset.fieldbyname(fmasterfield).AsVariant,[]);

     for id := 1 to dcnt do
     begin
      subcnt := 0;
      subno := 0;
      scnt := 0;

      doitd := fsubdetail.DataSet.Locate(fsubdetailfield,
         fdetail.dataset.fieldbyname(fsubmasterfield).AsVariant,[]);
      if doitd then
      while (fsubdetail.DataSet.fieldbyname(fsubdetailfield).asvariant =
            fdetail.dataset.fieldbyname(fsubmasterfield).AsVariant)
            and not fsubdetail.dataset.Eof do
      begin
       fsubdetail.DataSet.Next;
       inc(scnt);
      end;
      if not doitd and fshowemptysub then doitd := true;
      if doitd then
      begin
       inc(dtlgrpno);
       inc(dtlgrptotal);
       inc(dtlgrpcnt);

       Pvd.InsertRows(tkpos,allend);
       Pvd.CopyRows(dtpos,tkpos,allend);
       for j := 1 to dtlen do
       for k := 1 to ccnt do
       begin
        ArrayData[j,k] := dtFields[j-1][k-1];
         if dtfields[j-1][k-1] <> '' then
        begin
         if dtfields[j-1][k-1][1] = '&' then
         begin
          ArrayData[j,k] := getvalueex(dtfields[j-1][k-1],2);
          end
         else
         if dtfields[j-1][k-1][1] = '$' then
         begin
          nm := uppercase(copy(dtfields[j-1][k-1],2,length(dtfields[j-1][k-1])-1));
          split(nm,ds,s);
           case ds of
{$ifdef myvariant}
          'M': ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s));
          'D',' ': ArrayData[j,k] := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s));
          'S': ArrayData[j,k] := Pvd.GetMyVariant(fsubdetail.dataset.FieldbyName(s));
{$else}
          'M': ArrayData[j,k] := fmaster.dataset.FieldbyName(s).Value;//GetMyVariant(fmaster.dataset.FieldbyName(nm));
          'D',' ': ArrayData[j,k] := fdetail.dataset.FieldbyName(s).Value;//  GetMyVariant(fdetail.dataset.FieldbyName(nm));
          'S': ArrayData[j,k] := fsubdetail.dataset.FieldbyName(s).Value; // GetMyVariant(fsubdetail.dataset.FieldbyName(nm));
{$endif}
          else raise TArtRepException.Create('Wrong template!');
          end;
         end;
        end;
       end;
       if dtlen > 0 then Pvd.SetRange(tkpos,1,tkpos+dtlen - 1,
         ccnt, ArrayData);

       for j := 1 to dblen do
       for k := 1 to ccnt do
       begin
        ArrayData[j,k] := dbFields[j-1][k-1];
        if dbfields[j-1][k-1] <> '' then
         begin
         if dbfields[j-1][k-1][1] = '&' then
         begin
          ArrayData[j,k] := getvalueex(dbfields[j-1][k-1],2);
          end
         else
         if dbfields[j-1][k-1][1] = '$' then
         begin
          nm := uppercase(copy(dbfields[j-1][k-1],2,length(dbfields[j-1][k-1])-1));
          split(nm,ds,s);
          case ds of
{$ifdef myvariant}
          'M': ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s));
          'D',' ': ArrayData[j,k] := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s));
          'S': ArrayData[j,k] := Pvd.GetMyVariant(fsubdetail.dataset.FieldbyName(s));
{$else}
          'M': ArrayData[j,k] := fmaster.dataset.FieldbyName(s).Value;//  GetMyVariant(fmaster.dataset.FieldbyName(nm));
          'D',' ': ArrayData[j,k] := fdetail.dataset.FieldbyName(s).Value;//  GetMyVariant(fdetail.dataset.FieldbyName(nm));
          'S': ArrayData[j,k] := fsubdetail.dataset.FieldbyName(s).Value;//  GetMyVariant(fsubdetail.dataset.FieldbyName(nm));
{$endif}
          else raise TArtRepException.Create('Wrong template!');
          end;
         end;
        end;
       end;
       if dblen > 0 then Pvd.SetRange(tkpos+dbpos-dtpos,1,tkpos+dbpos-dtpos+dblen-1,
         ccnt, ArrayData);

       dttkpos := tkpos;
       inc(tkpos,dtlen);

// Sub
       if scnt = 0 then
       Pvd.DeleteRows(tkpos,slen)
       else
       begin
        fsubdetail.DataSet.Locate(fsubdetailfield,
            fdetail.dataset.fieldbyname(fsubmasterfield).AsVariant,[]);



        i2 := 1;
        k := scnt-1;
        while i2 < scnt do
        begin
         if k >= i2 then k1 := i2 else k1 := k;

         Pvd.InsertRows(tkpos+slen*k1,slen*k1);
         Pvd.CopyRows(tkpos,tkpos+slen*k1,slen*k1);
         inc(i2,k1);
         dec(k,k1);
        end;

        inc(subtotal,scnt);
        inc(suball,scnt);
        k1 := 0;
        k2 := 0;
        for i2 := 1 to scnt do
        begin
         subno := i2;
         inc(subcnt);
         inc(subtotalno);
         inc(suballno);
         for j := 1 to slen do
         for k := 1 to ccnt do
         begin
          ArrayData[k2+j,k] := sFields[j-1][k-1];
          if sfields[j-1][k-1] <> '' then
          begin
           if sfields[j-1][k-1][1] = '&' then
           begin
            ArrayData[k2+j,k] := getvalueex(sfields[j-1][k-1],3);
           end
           else
           if sfields[j-1][k-1][1] = '$' then
           begin
            nm := uppercase(copy(sfields[j-1][k-1],2,length(sfields[j-1][k-1])-1));
            split(nm,ds,s);
            case ds of
{$ifdef myvariant}
            'M': ArrayData[k2+j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(s));
            'D': ArrayData[k2+j,k] := Pvd.GetMyVariant(fdetail.dataset.FieldbyName(s));
            'S',' ': ArrayData[k2+j,k] := Pvd.GetMyVariant(fsubdetail.dataset.FieldbyName(s));
{$else}
            'M': ArrayData[k2+j,k] := fmaster.dataset.FieldbyName(s).Value;//  GetMyVariant(fmaster.dataset.FieldbyName(nm));
            'D': ArrayData[k2+j,k] := fdetail.dataset.FieldbyName(s).Value;//  GetMyVariant(fdetail.dataset.FieldbyName(nm));
            'S',' ': ArrayData[k2+j,k] := fsubdetail.dataset.FieldbyName(s).Value;//  GetMyVariant(fsubdetail.dataset.FieldbyName(nm));
{$endif}
            else raise TArtRepException.Create('Wrong template!');
            end;
           end;
          end;
         end;
         inc(k2,slen);
         fsubdetail.dataset.Next;
         inc(k1);
         if k1 mod farraysize = 0 then
         begin
          Pvd.SetRange(tkpos+(k1-farraysize)*slen,1,tkpos+k1*slen - 1,
             ccnt, ArrayData);
          k2 := 0;
         end;
        end;

        if k1 mod farraysize <> 0 then
        begin
          Pvd.SetRange(tkpos+(k1-k1 mod farraysize)*slen,1,tkpos+k1*slen - 1,
           ccnt, ArrayData);
        end;

       end;



       if fsubgrouping then
       begin
        AddGrp(tkpos,subcnt*slen);
       end;

       Pvd.AddNameRange(tkpos,1,tkpos+subcnt*slen - 1,1,fsubname+inttostr(masterno)+'_'+inttostr(dtlgrpno));
       Pvd.AddName('SubDetailCount'+inttostr(masterno)+'_'+inttostr(dtlgrpno),subcnt);

       inc(tkpos,scnt*slen);
       inc(tkpos,dblen);
      end;
      fdetail.DataSet.Next;
     end;

    end;

   Pvd.AddNameRange(mttkpos,1,tkpos - 1,1,fdetailname+inttostr(masterno));

   if fgrouping then
   begin
    AddGrp(mttkpos,tkpos - mttkpos);
   end;


   Pvd.AddName('DetailCount'+inttostr(masterno),dtlgrpcnt);
   Pvd.AddName('SubDetailTotal'+inttostr(masterno),subtotal);

   if tkpos = mttkpos then
   begin
     Pvd.DeleteRows(tkpos-mtlen,mblen+mtlen);
     dec(tkpos,mtlen);
   end
   else inc(tkpos,mblen);
   end;
   fmaster.DataSet.Next;
  end;

 Pvd.AddNameRange(mtpos,1,tkpos - 1,1,fmastername);

 for j:=0 to grpcount-1 do
  Pvd.Group(grpdata[j].row,grpdata[j].count);

 Pvd.DeleteRows(mtpos,allen);

 Pvd.AddName('MasterCount',mastercnt);
 Pvd.AddName('DetailTotal',dtlgrptotal);
 Pvd.AddName('SubDetailAll',suball);

 ArrayData := unassigned;

end;

procedure TArtReport.makegroup;
var dpos, mtpos, mbpos, allen, tkpos,
    mtlen, mblen, dlen, i,j,k,i2,k1,k2,
    dcnt, rcnt:integer;
    mtfields, dfields, mbfields : array of array of string;
    nm : string;
    ArrayData,Val  : Variant;
    Place: TBookmark;
begin

  grpcount := 0;
  dlen := pos[groupepos] - pos[groupbpos]+1;
  mtlen := pos[groupbpos] - pos[masterbpos];
  mblen := pos[masterepos] - pos[groupepos];
  allen := pos[masterepos] - pos[masterbpos]+1;
  dpos := pos[groupbpos];
  mtpos := pos[masterbpos];
  mbpos := pos[groupepos]+1;
  setlength(dfields,dlen,ccnt);
  if mtlen > 0 then setlength(mtfields,mtlen,ccnt);
  if mblen > 0 then setlength(mbfields,mblen,ccnt);

  for i := 0 to dlen-1 do
   for j := 1 to ccnt do
    dfIELDS[i,j-1] := Pvd.GetCellFormula(dpos+i,j);

  for i := 0 to mtlen-1 do
   for j := 1 to ccnt do
    mtfIELDS[i,j-1] := Pvd.GetCellFormula(mtpos+i,j);

  for i := 0 to mblen-1 do
   for j := 1 to ccnt do
    mbfIELDS[i,j-1] := Pvd.GetCellFormula(mbpos+i,j);

  ArrayData := VarArrayCreate([1,farraysize *dlen, 1, ccnt], varVariant);
  fmaster.dataset.First;
  tkpos := pos[masterepos]+1;

  dtlgrptotal := 0;
  dtlgrptotalno := 0;
  masterno := 0;
  mastercnt := 0;
  fmaster.dataset.first;
  rcnt := fmaster.dataset.recordcount;

//******************************************
  while not fmaster.dataset.eof do
  begin

   if assigned(fprogressproc) then
     fprogressproc(fmaster.dataset.recno,rcnt);

   dtlgrpcnt := 0;
   dtlgrpno := 0;
   dcnt := 0;
   Val := fmaster.DataSet.fieldbyname(fgroupfield).AsVariant;
   Place := fmaster.DataSet.GetBookmark;

   while (fmaster.DataSet.FieldByName(fgroupfield).AsVariant = Val)
         and not fmaster.DataSet.Eof do
   begin
    fmaster.DataSet.Next;
    inc(dcnt);
   end;

   fmaster.DataSet.GotoBookmark(Place);
   fmaster.DataSet.FreeBookmark(Place);
   inc(masterno);
   inc(mastercnt);


   Pvd.InsertRows(tkpos,allen);
   Pvd.CopyRows(mtpos,tkpos,allen);


   for j := 1 to mtlen do
   for k := 1 to ccnt do
    begin
     ArrayData[j,k] := mtFields[j-1][k-1];
     if mtfields[j-1][k-1] <> '' then
     begin
      if mtfields[j-1][k-1][1] = '&' then
      begin
       ArrayData[j,k] := getvalueex(mtfields[j-1][k-1]);
      end
      else
      if mtfields[j-1][k-1][1] = '$' then
      begin
       nm := uppercase(copy(mtfields[j-1][k-1],2,length(mtfields[j-1][k-1])-1));
{$ifdef myvariant}
       ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$else}
       ArrayData[j,k] := fmaster.dataset.FieldbyName(nm).Value;//GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$endif}
      end;
     end;
    end;
    if mtlen > 0 then Pvd.SetRange(tkpos,1,
       tkpos+mtlen - 1,ccnt, ArrayData);


    for j := 1 to mblen do
     for k := 1 to ccnt do
     begin
      ArrayData[j,k] := mbFields[j-1][k-1];
      if mbfields[j-1][k-1] <> '' then
      begin
       if mbfields[j-1][k-1][1] = '&' then
       begin
        ArrayData[j,k] := getvalueex(mbfields[j-1][k-1]);
       end
       else
       if mbfields[j-1][k-1][1] = '$' then
       begin
        nm := uppercase(copy(mbfields[j-1][k-1],2,length(mbfields[j-1][k-1])-1));
{$ifdef Myvariant}
        ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$else}
        ArrayData[j,k] := fmaster.dataset.FieldbyName(nm).Value;//GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$endif}
       end;
      end;
     end;
     if mblen > 0 then Pvd.SetRange(tkpos+mbpos-mtpos,1,tkpos+mbpos-mtpos+mblen-1,
        ccnt, ArrayData);

     inc(tkpos,mtlen);


      i2 := 1;
      k := dcnt-1;
      while i2 < dcnt do
      begin
       if k >= i2 then k1 := i2 else k1 := k;
       Pvd.InsertRows(tkpos+dlen*k1, dlen*k1);
       Pvd.CopyRows(tkpos,tkpos+dlen*k1,dlen*k1);
       inc(i2,k1);
       dec(k,k1);
      end;

      inc(dtlgrptotal,dcnt);
      k1 := 0;
      k2 := 0;
      for i2 := 1 to dcnt do
      begin
       dtlgrpno := i2;
       inc(dtlgrpcnt);
       inc(dtlgrptotalno);
       for j := 1 to dlen do
       for k := 1 to ccnt do
        begin
        ArrayData[k2+j,k] := dFields[j-1][k-1];
        if dfields[j-1][k-1] <> '' then
        begin
         if dfields[j-1][k-1][1] = '&' then
         begin
          ArrayData[k2+j,k] := getvalueex(dfields[j-1][k-1]);
         end
         else
         if dfields[j-1][k-1][1] = '$' then
         begin
          nm := uppercase(copy(dfields[j-1][k-1],2,length(dfields[j-1][k-1])-1));
{$ifdef myvariant}
          ArrayData[k2+j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$else}
          ArrayData[k2+j,k] := fmaster.dataset.FieldbyName(nm).Value;//GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$endif}
         end;
        end;
       end;
       inc(k2,dlen);
       fmaster.dataset.Next;
       inc(k1);
       if k1 mod farraysize = 0 then
       begin
        Pvd.SetRange(tkpos+(k1-farraysize)*dlen,1,tkpos+k1*dlen - 1,
           ccnt,ArrayData);
        k2 := 0;
       end;
      end;

      if k1 mod farraysize <> 0 then
      begin
         Pvd.SetRange(tkpos+(k1-k1 mod farraysize)*dlen,1,tkpos+k1*dlen - 1,
            ccnt, ArrayData);
      end;



     if fgrouping then
     begin
      AddGrp(tkpos,dtlgrpcnt*dlen);
     end;

     Pvd.AddNameRange(tkpos,1,tkpos+dtlgrpcnt*dlen - 1,1,fdetailname+inttostr(masterno));
     Pvd.AddName('GroupCount'+inttostr(masterno),dtlgrpcnt);

     inc(tkpos,dcnt*dlen);
     inc(tkpos,mblen);
  end;


  for j:=0 to grpcount-1 do Pvd.Group(grpdata[j].row,grpdata[j].count);
  Pvd.AddNameRange(mtpos,1,tkpos - 1,1,fmastername);
  Pvd.DeleteRows(mtpos, allen);
//  Pvd.AddName('MasterCount'+inttostr(masterno),inttostr(mastercnt));
//  Pvd.AddName('GroupTotal'+inttostr(masterno),inttostr(dtlgrptotal));
  Pvd.AddName('MasterCount',mastercnt);
  Pvd.AddName('GroupTotal',dtlgrptotal);

  ArrayData := unassigned;
  Val := unassigned;
end;


procedure TArtReport.makesubgroup;
var dtpos, dbpos, spos, mtpos, mbpos, allen, tkpos,
    mtlen, mblen, dtlen, dblen, slen, i,j,k,i2,k1,k2,id,
    allend, xpos, dttkpos, dcnt, scnt, mttkpos, rcnt:integer;
    mtfields, dtfields, mbfields, dbfields, sfields : array of array of string;
    s,nm : string;
    ArrayData,Val,Val2  : Variant;
    Place: TBookmark;
begin

  grpcount := 0;
  mtlen := pos[groupbpos] - pos[masterbpos];
  mblen := pos[masterepos] - pos[groupepos];
  dtlen := pos[subgroupbpos] - pos[groupbpos];
  dblen := pos[groupepos] - pos[subgroupepos];
  slen := pos[subgroupepos] - pos[subgroupbpos] + 1;
  allen := pos[masterepos] - pos[masterbpos]+1;
  allend := pos[groupepos] - pos[groupbpos]+1;
  spos := pos[subgroupbpos];
  mtpos := pos[masterbpos];
  mbpos := pos[groupepos]+1;
  dtpos := pos[groupbpos];
  dbpos := pos[subgroupepos]+1;

  setlength(sfields,slen,ccnt);
  if mtlen > 0 then setlength(mtfields,mtlen,ccnt);
  if mblen > 0 then setlength(mbfields,mblen,ccnt);
  if dtlen > 0 then setlength(dtfields,dtlen,ccnt);
  if dblen > 0 then setlength(dbfields,dblen,ccnt);

  for i := 0 to slen-1 do
   for j := 1 to ccnt do
    sfIELDS[i,j-1] := Pvd.GetCellFormula(spos+i,j);

  for i := 0 to mtlen-1 do
   for j := 1 to ccnt do
    mtfIELDS[i,j-1] := Pvd.GetCellFormula(mtpos+i,j);

  for i := 0 to mblen-1 do
   for j := 1 to ccnt do
    mbfIELDS[i,j-1] := Pvd.GetCellFormula(mbpos+i,j);

  for i := 0 to dtlen-1 do
   for j := 1 to ccnt do
    dtfIELDS[i,j-1] := Pvd.GetCellFormula(dtpos+i,j);

  for i := 0 to dblen-1 do
   for j := 1 to ccnt do
    dbfIELDS[i,j-1] := Pvd.GetCellFormula(dbpos+i,j);

  ArrayData := VarArrayCreate([1,farraysize *slen, 1, ccnt], varVariant);
  tkpos := pos[masterepos]+1;

  dtlgrptotal := 0;
  dtlgrptotalno := 0;
  masterno := 0;
  mastercnt := 0;
  suball := 0;
  suballno := 0;

// Master
  fmaster.DataSet.First;
  rcnt := fmaster.DataSet.recordcount;
  while not fmaster.dataset.eof do
  begin

   if assigned(fprogressproc) then
     fprogressproc(fmaster.dataset.recno,rcnt);

   dtlgrpcnt := 0;
   dcnt := 0;
   dtlgrpno := 0;
   subtotal := 0;
   subtotalno := 0;

   Val := fmaster.DataSet.fieldbyname(fgroupfield).AsVariant;
   Place := fmaster.DataSet.GetBookmark;

   while (fmaster.DataSet.FieldByName(fgroupfield).AsVariant = Val)
         and not fmaster.DataSet.Eof do
   begin
    inc(dcnt);
    Val2 := fmaster.DataSet.FieldByName(fsubgroupfield).AsVariant;
    while (fmaster.DataSet.FieldByName(fsubgroupfield).AsVariant = Val2) and
         (fmaster.DataSet.FieldByName(fgroupfield).AsVariant = Val) and
          not fmaster.DataSet.Eof do
     fmaster.DataSet.Next;
   end;

   fmaster.DataSet.GotoBookmark(Place);
   fmaster.DataSet.FreeBookmark(Place);

   inc(masterno);
   inc(mastercnt);

   Pvd.InsertRows(tkpos,allen);
   Pvd.CopyRows(mtpos,tkpos,allen);

   for j := 1 to mtlen do
    for k := 1 to ccnt do
    begin
     ArrayData[j,k] := mtFields[j-1][k-1];
     if mtfields[j-1][k-1] <> '' then
     begin
      if mtfields[j-1][k-1][1] = '&' then
      begin
       ArrayData[j,k] := getvalueex(mtfields[j-1][k-1]);
      end
      else
      if mtfields[j-1][k-1][1] = '$' then
      begin
       nm := uppercase(copy(mtfields[j-1][k-1],2,length(mtfields[j-1][k-1])-1));
{$ifdef myvariant}
       ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$else}
       ArrayData[j,k] := fmaster.dataset.FieldbyName(nm).Value;//  GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$endif}
      end;
     end;
    end;
   if mtlen > 0 then Pvd.SetRange(tkpos,1,tkpos+mtlen - 1,
       ccnt, ArrayData);


   for j := 1 to mblen do
    for k := 1 to ccnt do
    begin
     ArrayData[j,k] := mbFields[j-1][k-1];
     if mbfields[j-1][k-1] <> '' then
     begin
      if mbfields[j-1][k-1][1] = '&' then
      begin
       ArrayData[j,k] := getvalueex(mbfields[j-1][k-1]);
      end
      else
      if mbfields[j-1][k-1][1] = '$' then
      begin
       nm := uppercase(copy(mbfields[j-1][k-1],2,length(mbfields[j-1][k-1])-1));
{$ifdef myvariant}
       ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$else}
       ArrayData[j,k] := fmaster.dataset.FieldbyName(nm).Value;
{$endif}
      end;
     end;
//     MessageBox(0,pchar(VarToStr(ArrayData[j,k])),'',0);
    end;
   if mblen > 0 then Pvd.SetRange(tkpos+mbpos-mtpos,1,tkpos+mbpos-mtpos+mblen-1,
                     ccnt, ArrayData);

   inc(tkpos,mtlen);
   mttkpos := tkpos;
// Detail
   Pvd.DeleteRows(tkpos, allend);

   for id := 1 to dcnt do
   begin
      subcnt := 0;
      subno := 0;
      scnt := 0;

      Val2 := fmaster.DataSet.fieldbyname(fsubgroupfield).AsVariant;
      Val := fmaster.DataSet.fieldbyname(fgroupfield).AsVariant;
      Place := fmaster.DataSet.GetBookmark;

      while (fmaster.DataSet.FieldByName(fsubgroupfield).AsVariant = Val2) and
            (fmaster.DataSet.FieldByName(fgroupfield).AsVariant = Val) and
            not fmaster.DataSet.Eof do
      begin
       fmaster.DataSet.Next;
       inc(scnt);
      end;

      fmaster.DataSet.GotoBookmark(Place);
      fmaster.DataSet.FreeBookmark(Place);

      inc(dtlgrpno);
      inc(dtlgrptotal);
      inc(dtlgrpcnt);

      Pvd.InsertRows(tkpos,allend);
      Pvd.CopyRows(dtpos,tkpos,allend);
      for j := 1 to dtlen do
       for k := 1 to ccnt do
       begin
        ArrayData[j,k] := dtFields[j-1][k-1];
         if dtfields[j-1][k-1] <> '' then
        begin
         if dtfields[j-1][k-1][1] = '&' then
         begin
          ArrayData[j,k] := getvalueex(dtfields[j-1][k-1]);
          end
         else
         if dtfields[j-1][k-1][1] = '$' then
         begin
          nm := uppercase(copy(dtfields[j-1][k-1],2,length(dtfields[j-1][k-1])-1));
{$ifdef myvariant}
          ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$else}
          ArrayData[j,k] := fmaster.dataset.FieldbyName(nm).Value;//  GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$endif}
         end;
        end;
       end;
      if dtlen > 0 then Pvd.SetRange(tkpos,1,tkpos+dtlen - 1, ccnt, ArrayData);


      for j := 1 to dblen do
       for k := 1 to ccnt do
       begin
        ArrayData[j,k] := dbFields[j-1][k-1];
        if dbfields[j-1][k-1] <> '' then
         begin
         if dbfields[j-1][k-1][1] = '&' then
         begin
          ArrayData[j,k] := getvalueex(dbfields[j-1][k-1]);
          end
         else
         if dbfields[j-1][k-1][1] = '$' then
         begin
          nm := uppercase(copy(dbfields[j-1][k-1],2,length(dbfields[j-1][k-1])-1));
{$ifdef myvariant}
          ArrayData[j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$else}
          ArrayData[j,k] := fmaster.dataset.FieldbyName(nm).Value;//  GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$endif}
         end;
        end;
       end;
      if dblen > 0 then Pvd.SetRange(tkpos+dbpos-dtpos,1,tkpos+dbpos-dtpos+dblen-1,ccnt, ArrayData);

      dttkpos := tkpos;
      inc(tkpos,dtlen);

// Sub

      i2 := 1;
      k := scnt-1;
      while i2 < scnt do
      begin
         if k >= i2 then k1 := i2 else k1 := k;
         Pvd.InsertRows(tkpos+slen*k1, slen*k1);
         Pvd.CopyRows(tkpos, tkpos+slen*k1, slen*k1);
         inc(i2,k1);
         dec(k,k1);
      end;

      inc(subtotal,scnt);
      inc(suball,scnt);
      k1 := 0;
      k2 := 0;
      for i2 := 1 to scnt do
      begin
         subno := i2;
         inc(subcnt);
         inc(subtotalno);
         inc(suballno);
         for j := 1 to slen do
         for k := 1 to ccnt do
         begin
          ArrayData[k2+j,k] := sFields[j-1][k-1];
          if sfields[j-1][k-1] <> '' then
          begin
           if sfields[j-1][k-1][1] = '&' then
           begin
            ArrayData[k2+j,k] := getvalueex(sfields[j-1][k-1]);
           end
           else
           if sfields[j-1][k-1][1] = '$' then
           begin
            nm := uppercase(copy(sfields[j-1][k-1],2,length(sfields[j-1][k-1])-1));
{$ifdef myvariant}
            ArrayData[k2+j,k] := Pvd.GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$else}
            ArrayData[k2+j,k] := fmaster.dataset.FieldbyName(nm).Value;//  GetMyVariant(fmaster.dataset.FieldbyName(nm));
{$endif}
           end;
          end;
         end;
         inc(k2,slen);
         fmaster.dataset.Next;
         inc(k1);
         if k1 mod farraysize = 0 then
         begin
          Pvd.SetRange(tkpos+(k1-farraysize)*slen,1,tkpos+k1*slen - 1,
              ccnt, ArrayData);
          k2 := 0;
         end;
      end;

      if k1 mod farraysize <> 0 then
      begin
        Pvd.SetRange(tkpos+(k1-k1 mod farraysize)*slen,1,tkpos+k1*slen - 1,
             ccnt, ArrayData);
      end;



      if fsubgrouping then
      begin
       AddGrp(tkpos, subcnt*slen);
      end;

      Pvd.AddNameRange(tkpos,1,tkpos+subcnt*slen - 1,1,fsubname+inttostr(masterno)+'_'+inttostr(dtlgrpno));
      Pvd.AddName('SubGroupCount'+inttostr(masterno)+'_'+inttostr(dtlgrpno),subcnt);

      inc(tkpos,scnt*slen);
      inc(tkpos,dblen);
   end;

   Pvd.AddNameRange(mttkpos,1,tkpos - 1,1,fdetailname+inttostr(masterno));

   if fgrouping then
   begin
    AddGrp(mttkpos,tkpos - mttkpos);
   end;

   Pvd.AddName('GroupCount'+inttostr(masterno),dtlgrpcnt);
   Pvd.AddName('SubGroupTotal'+inttostr(masterno),subtotal);

   inc(tkpos,mblen);
  end;

 Pvd.AddNameRange(mtpos,1,tkpos - 1,1,fmastername);

 for j:=0 to grpcount-1 do Pvd.Group(grpdata[j].row,grpdata[j].count);

 Pvd.DeleteRows(mtpos, allen);

 Pvd.AddName('MasterCount',mastercnt);
 Pvd.AddName('GroupTotal',dtlgrptotal);
 Pvd.AddName('SubGroupAll',suball);

 ArrayData := unassigned;
 Val := unassigned;
 Val2 := unassigned;

end;


procedure TArtReport.make;
var i,j,rcnt:integer;
    s : string;
    x : TXPos;
begin
 try
  checkmulticolumn(ftemplate);
 except
  raise;
 end;
 settemplate;
 if ftype = rtnone then raise TArtRepException.Create('Wrong template!');
 try
  rcnt := Pvd.LastUsedRow;
  ccnt := Pvd.LastUsedColumn;
  setlength(autofit,ccnt);
  Pvd.DeleteRow(1);
  for i := 1 to ccnt do
   autofit[i-1] := uppercase(Pvd.GetCellText(1,i)) = '#AUTOFIT';
  Pvd.DeleteRow(1);

  dec(rcnt,2);
  i := 1;

  for x := beginpos to subgroupepos do
   pos[x] := 0;

  while i <= rcnt do
  begin
   s := lowercase(Pvd.GetCellText(i,1));
   if s = '#begin' then
   begin
    if pos[beginpos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[beginpos] := i;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#end' then
   begin
    if pos[endpos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[endpos] := i-1;
     Pvd.DeleteRow(i);
     break;
    end;
   end
   else
   if s = '#masterbegin' then
   begin
    if pos[masterbpos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[masterbpos] := i;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#masterend' then
   begin
    if pos[masterepos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[masterepos] := i-1;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#detailbegin' then
   begin
    if pos[detailbpos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[detailbpos] := i;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#detailend' then
   begin
    if pos[detailepos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[detailepos] := i-1;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#subdetailbegin' then
   begin
    if pos[subdetailbpos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[subdetailbpos] := i;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#subdetailend' then
   begin
    if pos[subdetailepos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[subdetailepos] := i-1;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#groupbegin' then
   begin
    if pos[groupbpos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[groupbpos] := i;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#groupend' then
   begin
    if pos[groupepos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[groupepos] := i-1;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#subgroupbegin' then
   begin
    if pos[subgroupbpos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[subgroupbpos] := i;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else
   if s = '#subgroupend' then
   begin
    if pos[subgroupepos] <> 0 then raise TArtRepException.Create('Wrong template!')
    else
    begin
     pos[subgroupepos] := i-1;
     dec(rcnt);
     Pvd.DeleteRow(i);
    end;
   end
   else inc(i);
  end;

  if (pos[beginpos] = 0) or (pos[endpos] = 0) or
     (pos[masterbpos] = 0) or (pos[masterepos] = 0) or
     (pos[beginpos] > pos[endpos]) or
     (pos[masterbpos] > pos[masterepos]) or
     (pos[beginpos] > pos[masterbpos]) or
     (pos[endpos] < pos[masterepos]) then
  raise TArtRepException.Create('Wrong template!');

  case ftype of
  rtsimple:
  begin
   if (pos[detailbpos] <> 0) or (pos[detailepos] <> 0) or
      (pos[subdetailbpos] <> 0) or (pos[subdetailepos] <> 0) or
      (pos[groupbpos] <> 0) or (pos[groupepos] <> 0) or
      (pos[subgroupbpos] <> 0) or (pos[subgroupepos] <> 0) then
   raise TArtRepException.Create('Wrong template!');
   if (fmaster = nil) or (fmaster.DataSet = nil) then
    raise TArtRepException.Create('Master DataSet not set!');
  end;
  rtdetail:
  begin
   if (pos[detailbpos] = 0) or (pos[detailepos] = 0) or
      (pos[subdetailbpos] <> 0) or (pos[subdetailepos] <> 0) or
      (pos[groupbpos] <> 0) or (pos[groupepos] <> 0) or
      (pos[subgroupbpos] <> 0) or (pos[subgroupepos] <> 0) or
      (pos[detailbpos] > pos[detailepos]) or
      (pos[beginpos] > pos[detailbpos]) or
      (pos[detailepos] > pos[endpos]) then
   raise TArtRepException.Create('Wrong template!');
   if (fmaster = nil) or (fmaster.DataSet = nil) then
    raise TArtRepException.Create('Master DataSet not set!');
   if (fdetail = nil) or (fdetail.DataSet = nil) then
    raise TArtRepException.Create('Detail DataSet not set!');
   if fmasterfield = '' then
    raise TArtRepException.Create('MasterField not set!');
   if fdetailfield = '' then
    raise TArtRepException.Create('DetailField not set!');
  end;
  rtsubdetail:
  begin
   if (pos[detailbpos] = 0) or (pos[detailepos] = 0) or
      (pos[subdetailbpos] = 0) or (pos[subdetailepos] = 0) or
      (pos[groupbpos] <> 0) or (pos[groupepos] <> 0) or
      (pos[subgroupbpos] <> 0) or (pos[subgroupepos] <> 0) or
      (pos[detailbpos] > pos[detailepos]) or
      (pos[beginpos] > pos[detailbpos]) or
      (pos[detailepos] > pos[endpos]) or
      (pos[subdetailbpos] > pos[subdetailepos]) or
      (pos[detailbpos] > pos[subdetailbpos]) or
      (pos[subdetailepos] > pos[detailepos]) then
   raise TArtRepException.Create('Wrong template!');
   if (fmaster = nil) or (fmaster.DataSet = nil) then
    raise TArtRepException.Create('Master DataSet not set!');
   if (fdetail = nil) or (fdetail.DataSet = nil) then
    raise TArtRepException.Create('Detail DataSet not set!');
   if (fsubdetail = nil) or (fsubdetail.DataSet = nil) then
    raise TArtRepException.Create('SubDetail DataSet not set!');
   if fmasterfield = '' then
    raise TArtRepException.Create('MasterField not set!');
   if fdetailfield = '' then
    raise TArtRepException.Create('DetailField not set!');
   if fsubmasterfield = '' then
    raise TArtRepException.Create('SubMasterField not set!');
   if fsubdetailfield = '' then
    raise TArtRepException.Create('SubDetailField not set!');
  end;
  rtgroup:
  begin
   if (pos[detailbpos] <> 0) or (pos[detailepos] <> 0) or
      (pos[subdetailbpos] <> 0) or (pos[subdetailepos] <> 0) or
      (pos[groupbpos] = 0) or (pos[groupepos] = 0) or
      (pos[subgroupbpos] <> 0) or (pos[subgroupepos] <> 0) or
      (pos[groupbpos] > pos[groupepos]) or
      (pos[beginpos] > pos[groupbpos]) or
      (pos[groupepos] > pos[endpos]) then
   raise TArtRepException.Create('Wrong template!');
   if (fmaster = nil) or (fmaster.DataSet = nil) then
    raise TArtRepException.Create('Master DataSet not set!');
   if fgroupfield = '' then
    raise TArtRepException.Create('GroupField not set!');
  end;
  rtsubgroup:
  begin
   if (pos[detailbpos] <> 0) or (pos[detailepos] <> 0) or
      (pos[subdetailbpos] <> 0) or (pos[subdetailepos] <> 0) or
      (pos[groupbpos] = 0) or (pos[groupepos] = 0) or
      (pos[subgroupbpos] = 0) or (pos[subgroupepos] = 0) or
      (pos[groupbpos] > pos[groupepos]) or
      (pos[beginpos] > pos[groupbpos]) or
      (pos[groupepos] > pos[endpos]) or
      (pos[subgroupbpos] > pos[subgroupepos]) or
      (pos[groupbpos] > pos[subgroupbpos]) or
      (pos[subgroupepos] > pos[groupepos]) then
   raise TArtRepException.Create('Wrong template!');
   if (fmaster = nil) or (fmaster.DataSet = nil) then
    raise TArtRepException.Create('Master DataSet not set!');
   if fgroupfield = '' then
    raise TArtRepException.Create('GroupField not set!');
   if fsubgroupfield = '' then
    raise TArtRepException.Create('SubGroupField not set!');
  end
  end;

  fmaster.DataSet.active := true;
  if (ftype = rtdetail) or (ftype = rtsubdetail) then
   fdetail.DataSet.active := true;
  if ftype = rtsubdetail then
   fsubdetail.DataSet.active := true;

  for i := pos[beginpos] to pos[masterbpos]-1 do
   for j := 1 to ccnt do
   begin
    s := Pvd.GetCellFormula(i,j);
    if (length(s) > 1) then
     if(s[1] = '&') then
      Pvd.SetCellFormula(i,j,getvalue(s));
   end;

  for i := pos[masterepos]+1 to pos[endpos] do
   for j := 1 to ccnt do
   begin
    s := Pvd.GetCellFormula(i,j);
    if (s <> '') then
     if(s[1] = '&') then
      Pvd.SetCellFormula(i,j,getvalue(s));
   end;

   case ftype of
   rtsimple:
   begin
    fmaster.DataSet.DisableControls;
    makesimple;
    fmaster.DataSet.EnableControls;
   end;
   rtdetail:
   begin
    fmaster.DataSet.DisableControls;
    fdetail.DataSet.DisableControls;
    makedetail;
    fmaster.DataSet.EnableControls;
    fdetail.DataSet.EnableControls;
   end;
   rtsubdetail:
   begin
    fmaster.DataSet.DisableControls;
    fdetail.DataSet.DisableControls;
    fsubdetail.DataSet.DisableControls;
    makesubdetail;
    fmaster.DataSet.EnableControls;
    fdetail.DataSet.EnableControls;
    fsubdetail.DataSet.EnableControls;
   end;
   rtgroup:
   begin
    fmaster.DataSet.DisableControls;
    makegroup;
    fmaster.DataSet.EnableControls;
   end;
   rtsubgroup:
   begin
    fmaster.DataSet.DisableControls;
    makesubgroup;
    fmaster.DataSet.EnableControls;
   end;
   end;

  Pvd.CleanUp;

  for i := 1 to ccnt do
  begin
   if autofit[i-1] then Pvd.AutoFit(i);
  end;

  Pvd.ActivateCell(1,1);
//  (Wbk.Sheets.item[1] as _WorkSheet).Activate(lcid);


  if fsubgrouping then
      Pvd.ShowLevels(2);
  if fgrouping then
      Pvd.ShowLevels(1);

  if fsheetname <> '' then Pvd.SetSheetName(fsheetname);

  if fmacro <> '' then Pvd.Run(fmacro);


   if fpage = 0 then
    begin
     if fsaveas = '' then
     begin
      s := ftemplate;
      s := changefileext( {$IFDEF FPC} ExpandFileNameUTF8(s) { *Converted from ExpandFileName*  }
                               {$ELSE} ExpandFileName(s) {$ENDIF},'.xls');
     end
     else s := {$IFDEF FPC} ExpandFileNameUTF8(fsaveas); { *Converted from ExpandFileName*  }
                    {$ELSE} ExpandFileName(fsaveas); {$ENDIF}

     if {$IFDEF FPC} FileExistsUTF8(s) { *Converted from FileExists*  } then DeleteFileUTF8(s); { *Converted from DeleteFile*  }
             {$ELSE} FileExists(s) then DeleteFile(@s[1]); {$ENDIF}
      Pvd.SaveDocAs(s);
    end
    else
    begin
     s := ftemplate;
     Pvd.SaveDoc;
    end;

    Pvd.CloseDoc;
    if fprint then
    begin
     Pvd.OpenDoc(s,fpage);
     Pvd.Print;
     Pvd.CloseDoc;
    end; 

 except

  Pvd.BeforeQuit;

  if Pvd.DocLoaded then
  begin
   Pvd.MarkSaved;
   Pvd.CloseDoc;
  end;
  Pvd.Quit;
  raise;
 end;
  Pvd.BeforeQuit;
  Pvd.Quit;
end;


function TArtReport.IntReport:integer;
var
  s : string;
begin
 try
  result := 0;

  if not Pvd.Installed then
   raise TArtRepException.Create(Pvd.ProviderName+' not installed!');

  make;

  if fsaveas = '' then
  begin
   s := ftemplate;

   {$IFDEF FPC}
   s := changefileext(ExpandFileNameUTF8(s) { *Converted from ExpandFileName*  },'.xls');
  end
  else s := ExpandFileNameUTF8(fsaveas); { *Converted from ExpandFileName*  }
   if fshowreport then if FileExistsUTF8(s) { *Converted from FileExists*  } then
   {$ELSE}
   s := changefileext(ExpandFileName(s),'.xls');
  end
  else s := ExpandFileName(fsaveas);
   if fshowreport then if FileExists(s) then
   {$ENDIF}

   shellexecute(0,'open',@s[1],'','',sw_showmaximized);
 except
  on E:exception do
  begin
   ferror := E.Message;
   result := -1;
  end;
 end;
end;


{$IFDEF FPC}
initialization
     {$I tartreport.lrs}
{$IFDEF ARTXLSREPORT}
     {$I tartxlsreport.lrs}
{$ENDIF}     
{$ENDIF}

end.
